from schemey.factory.any_of_schema_factory import AnyOfSchemaFactory
from schemey.factory.array_schema_factory import ArraySchemaFactory
from schemey.factory.dataclass_schema_factory import DataclassSchemaFactory
from schemey.factory.datetime_factory import DatetimeFactory
from schemey.factory.enum_schema_factory import EnumSchemaFactory
from schemey.factory.external_type_factory import ExternalTypeFactory
from schemey.factory.factory_schema_factory import FactorySchemaFactory
from schemey.factory.impl_schema_factory import ImplSchemaFactory
from schemey.factory.ref_schema_factory import RefSchemaFactory
from schemey.factory.simple_type_factory import SimpleTypeFactory
from schemey.factory.tuple_schema_factory import TupleSchemaFactory
from schemey.factory.uuid_factory import UuidFactory
from schemey.schema_context import SchemaContext

priority = 100


def configure(context: SchemaContext):
    context.register_factory(RefSchemaFactory())
    context.register_factory(SimpleTypeFactory(bool, "boolean"))
    context.register_factory(SimpleTypeFactory(int, "integer"))
    context.register_factory(SimpleTypeFactory(type(None), "null"))
    context.register_factory(SimpleTypeFactory(float, "number"))
    context.register_factory(SimpleTypeFactory(str, "string"))
    context.register_factory(DatetimeFactory())
    context.register_factory(UuidFactory())
    context.register_factory(ArraySchemaFactory())
    context.register_factory(TupleSchemaFactory())
    context.register_factory(ExternalTypeFactory())
    context.register_factory(DataclassSchemaFactory())
    context.register_factory(EnumSchemaFactory())
    context.register_factory(FactorySchemaFactory())
    context.register_factory(ImplSchemaFactory())
    context.register_factory(AnyOfSchemaFactory())



from typing import Dict, Optional, Type

from marshy.types import ExternalItemType

from schemey import SchemaContext, Schema
from schemey.factory.schema_factory_abc import SchemaFactoryABC


class RefSchemaFactory(SchemaFactoryABC):
    """Schema factory for resolving json refs."""

    priority: int = 220

    def from_type(
        self, type_: Type, context: SchemaContext, path: str
    ) -> Optional[Schema]:
        pass

    def from_json(
        self,
        item: ExternalItemType,
        context: SchemaContext,
        path: str,
        ref_schemas: Dict[str, Schema],
    ) -> Optional[Schema]:
        ref = item.get("$ref")
        if ref:
            return ref_schemas.get(ref)



from enum import Enum


class StringFormat(Enum):
    DATE_TIME = "date-time"
    TIME = "time"
    DATE = "date"
    EMAIL = "email"
    HOSTNAME = "hostname"
    IPV4 = "ipv4"
    IPV6 = "ipv6"
    UUID = "uuid"
    URI = "uri"



from typing import Type, Optional, Set, List, Tuple, Union, Dict

import typing_inspect
from marshy.types import ExternalItemType
from marshy.utils import resolve_forward_refs

from schemey.factory.schema_factory_abc import SchemaFactoryABC
from schemey.schema import Schema
from schemey.schema_context import SchemaContext


class ArraySchemaFactory(SchemaFactoryABC):
    def from_type(
        self, type_: Type, context: SchemaContext, path: str
    ) -> Optional[Schema]:
        array_type = self.get_array_type(type_)
        if array_type:
            schema = {"type": "array"}
            args = typing_inspect.get_args(type_)
            if args:
                item_type = resolve_forward_refs(args[0])
                if item_type and not typing_inspect.is_typevar(item_type):
                    item_schema = context.schema_from_type(item_type, f"{path}/items")
                    schema["items"] = item_schema.schema
            if array_type is Set:
                schema["uniqueItems"] = True
            elif array_type is Tuple:
                schema["tuple"] = True  # Custom annotation
            return Schema(schema, type_)

    def from_json(
        self,
        item: ExternalItemType,
        context: SchemaContext,
        path: str,
        ref_schemas: Dict[str, Schema],
    ) -> Optional[Schema]:
        if item.get("type") == "array":
            type_ = List
            if item.get("uniqueItems") is True:
                type_ = Set
            elif item.get("tuple") is True:  # Custom annotation
                type_ = Tuple
            items = item.get("items")
            if not items:
                return Schema(item, type_)
            # noinspection PyTypeChecker
            item_schema = context.schema_from_json(items, f"{path}/items", ref_schemas)
            if type_ is Tuple:
                # noinspection PyTypeChecker
                python_type = type_[item_schema.python_type, ...]
            else:
                # noinspection PyTypeChecker
                python_type = type_[item_schema.python_type]
            return Schema(item, python_type)

    @staticmethod
    def get_array_type(type_: Type) -> Union[Type[List], Type[Set], Type[Tuple], None]:
        origin = typing_inspect.get_origin(type_)
        if origin is list:
            return List
        if origin is set:
            return Set
        if origin is tuple:
            args = typing_inspect.get_args(type_)
            if len(args) == 2 and args[1] is Ellipsis:
                return Tuple



from dataclasses import dataclass
from typing import Optional, Type, Dict

from marshy.types import ExternalItemType

from schemey.factory.schema_factory_abc import SchemaFactoryABC
from schemey.schema import Schema
from schemey.schema_context import SchemaContext


@dataclass
class SimpleTypeFactory(SchemaFactoryABC):
    python_type: Type
    json_type: str
    priority: int = 200

    def from_type(
        self, type_: Type, context: SchemaContext, path: str
    ) -> Optional[Schema]:
        if type_ == self.python_type:
            return Schema({"type": self.json_type}, type_)

    def from_json(
        self,
        item: ExternalItemType,
        context: SchemaContext,
        path: str,
        ref_schemas: Dict[str, Schema],
    ) -> Optional[Schema]:
        type_ = item.get("type")
        if type_ == self.json_type:
            return Schema(item, self.python_type)



from datetime import datetime

from jsonschema import FormatChecker


class SchemeyFormatChecker(FormatChecker):
    """
    As of right now, the standard jsonschema format checker does not do date-time, so we add it
    """

    def __init__(self):
        super().__init__()
        # noinspection PyTypeChecker
        self.checkers["date-time"] = (is_datetime, ValueError)


def is_datetime(value: str):
    return datetime.fromisoformat(value)



from dataclasses import dataclass
from typing import Type, Optional

from marshy import ExternalType
from marshy.marshaller.marshaller_abc import MarshallerABC

from schemey import Schema, get_default_schema_context, SchemaContext


@dataclass(frozen=True)
class SchemaMarshaller(MarshallerABC[Schema]):
    marshalled_type: Type[Schema] = Schema
    context: Optional[SchemaContext] = None

    def load(self, item: ExternalType) -> Schema:
        context = self.context or get_default_schema_context()
        loaded = context.schema_from_json(item)
        return loaded

    def dump(self, item: Schema) -> ExternalType:
        return item.schema



from dataclasses import dataclass
from enum import Enum
from inspect import isclass
from typing import Type, Optional, Dict

from marshy.types import ExternalItemType

from schemey.factory.schema_factory_abc import SchemaFactoryABC
from schemey.schema import Schema
from schemey.schema_context import SchemaContext


@dataclass
class EnumSchemaFactory(SchemaFactoryABC):
    priority: int = 190

    def from_type(
        self, type_: Type, context: SchemaContext, path: str
    ) -> Optional[Schema]:
        if isclass(type_) and issubclass(type_, Enum):
            # noinspection PyTypeChecker
            schema = dict(name=type_.__name__, enum=[e.value for e in type_])
            return Schema(schema, type_)

    def from_json(
        self,
        item: ExternalItemType,
        context: SchemaContext,
        path: str,
        ref_schemas: Dict[str, Schema],
    ) -> Optional[Schema]:
        name = item.get("name")
        enum = item.get("enum")
        if name and enum and isinstance(enum, list):
            if next((False for e in enum if not isinstance(e, str)), True):
                type_ = Enum(name, {e: e for e in enum})
                return Schema(item, type_)



from dataclasses import dataclass
from datetime import datetime
from typing import Type, Iterator, Optional, Union
from uuid import UUID

from jsonschema import ValidationError
from jsonschema.validators import validator_for
from marshy.types import ExternalItemType, ExternalType

from schemey.schemey_format_checker import SchemeyFormatChecker
from schemey.string_format import StringFormat
from schemey.util import filter_none


@dataclass
class Schema:
    schema: ExternalItemType
    python_type: Type

    def validator(self):
        validator = validator_for(self.schema)(
            schema=self.schema, format_checker=SchemeyFormatChecker()
        )
        return validator

    def iter_errors(self, item: ExternalType) -> Iterator[ValidationError]:
        yield from self.validator().iter_errors(instance=item)

    def validate(self, item: ExternalType):
        self.validator().validate(instance=item)


def str_schema(
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    str_format: Union[type(None), str, StringFormat] = None,
    pattern: Optional[str] = None,
) -> Schema:
    if isinstance(str_format, str):
        str_format = StringFormat(str_format)
    if str_format:
        str_format = str_format.value
    return Schema(
        filter_none(
            {
                "type": "string",
                "minLength": min_length,
                "maxLength": max_length,
                "format": str_format,
                "pattern": pattern,
            }
        ),
        str,
    )


def int_schema(
    minimum: Optional[int] = None,
    maximum: Optional[int] = None,
    exclusive_minimum: Optional[int] = None,
    exclusive_maximum: Optional[int] = None,
) -> Schema:
    return Schema(
        filter_none(
            {
                "type": "integer",
                "minimum": minimum,
                "maximum": maximum,
                "exclusiveMinimum": exclusive_minimum,
                "exclusiveMaximum": exclusive_maximum,
            }
        ),
        int,
    )


def float_schema(
    minimum: Optional[float] = None,
    maximum: Optional[float] = None,
    exclusive_minimum: Optional[float] = None,
    exclusive_maximum: Optional[float] = None,
) -> Schema:
    return Schema(
        filter_none(
            {
                "type": "number",
                "minimum": minimum,
                "maximum": maximum,
                "exclusiveMinimum": exclusive_minimum,
                "exclusiveMaximum": exclusive_maximum,
            }
        ),
        float,
    )


def uuid_schema():
    return Schema({"type": "string", "format": "uuid"}, UUID)


def datetime_schema():
    return Schema({"type": "string", "format": "date-time"}, datetime)



from marshy.marshaller_context import MarshallerContext

from schemey.schema_marshaller import SchemaMarshaller

priority = 100


def configure(context: MarshallerContext):
    context.register_marshaller(SchemaMarshaller())



from typing import Type, Optional, Union, Dict

import typing_inspect
from marshy.types import ExternalItemType

from schemey.factory.schema_factory_abc import SchemaFactoryABC
from schemey.schema import Schema
from schemey.schema_context import SchemaContext


class AnyOfSchemaFactory(SchemaFactoryABC):
    priority: int = 110

    def from_type(
        self, type_: Type, context: SchemaContext, path: str
    ) -> Optional[Schema]:
        origin = typing_inspect.get_origin(type_)
        if origin == Union:
            args = typing_inspect.get_args(type_)
            schemas = {
                "anyOf": [
                    context.schema_from_type(a, f"{path}/anyOf/{i}").schema
                    for i, a in enumerate(args)
                ]
            }
            return Schema(schemas, type_)

    def from_json(
        self,
        item: ExternalItemType,
        context: SchemaContext,
        path: str,
        ref_schemas: Dict[str, Schema],
    ) -> Optional[Schema]:
        any_of = item.get("anyOf")
        if any_of:
            schemas = (
                context.schema_from_json(a, f"{path}/anyOf/{i}", ref_schemas)
                for i, a in enumerate(any_of)
            )
            union = Union[tuple(s.python_type for s in schemas)]
            return Schema(item, union)



from dataclasses import dataclass
from typing import Type, Optional, Set, Dict

from marshy.factory.impl_marshaller_factory import ImplMarshallerFactory
from marshy.types import ExternalItemType

from schemey.factory.schema_factory_abc import SchemaFactoryABC
from schemey.schema import Schema
from schemey.schema_context import SchemaContext


@dataclass
class ImplSchemaFactory(SchemaFactoryABC):
    """
    Schema factory which generates schemas for base classes with implementations
    as set up in marshy. There is no facility for turning this back into a class
    structure - though the UnionFactory will make a reasonably standardized class
    structure from the result.
    """

    priority: int = 150

    def from_type(
        self, type_: Type, context: SchemaContext, path: str
    ) -> Optional[Schema]:
        impls = self.get_impls(type_, context)
        if impls:
            impls = sorted(list(impls), key=lambda i: i.__name__)
            any_of = []
            for impl in impls:
                prefix_items = [
                    {"const": impl.__name__},
                    context.schema_from_type(
                        impl, f"{path}/anyOf/{len(any_of)}/prefixItems/1"
                    ).schema,
                ]
                any_of.append(
                    {"type": "array", "prefixItems": prefix_items, "items": False}
                )
            return Schema({"anyOf": any_of}, type_)

    def from_json(
        self,
        item: ExternalItemType,
        context: SchemaContext,
        path: str,
        ref_schemas: Dict[str, Schema],
    ) -> Optional[Schema]:
        """No implementation since it is unlikely to be needed"""

    @staticmethod
    def get_impls(type_: Type, context: SchemaContext) -> Optional[Set[Type]]:
        factories = context.marshaller_context.get_factories()
        for factory in factories:
            if isinstance(factory, ImplMarshallerFactory):
                if factory.base == type_:
                    return factory.impls



import dataclasses
from typing import Type, Optional, get_type_hints, Dict

import typing_inspect
from marshy.types import ExternalItemType

from schemey.factory.schema_factory_abc import SchemaFactoryABC
from schemey.schema import Schema
from schemey.schema_context import SchemaContext


class DataclassSchemaFactory(SchemaFactoryABC):
    def from_type(
        self, type_: Type, context: SchemaContext, path: str
    ) -> Optional[Schema]:
        if not dataclasses.is_dataclass(type_):
            return
        # Setting this here will mean any nested references will not fail
        schema = Schema({"$ref": "/".join(path)}, type_)
        context.schemas_by_type[type_] = schema
        # noinspection PyDataclass
        fields = dataclasses.fields(type_)
        try:
            types = get_type_hints(type_, globalns=None, localns=None)
        except NameError:
            types = {f.name: f.type for f in fields}

        properties = {}
        required = []
        for field in fields:
            field_schema = field.metadata.get("schemey")
            if not field_schema:
                field_schema = context.schema_from_type(
                    types[field.name], f"{path}/properties/{field.name}"
                )
            if (
                field.default is dataclasses.MISSING
                and field.default_factory is dataclasses.MISSING
            ):
                required.append(field.name)
            if field.default is not dataclasses.MISSING:
                field_schema = Schema({**field_schema.schema}, field_schema.python_type)
                field_schema.schema["default"] = context.marshaller_context.dump(
                    field.default, types[field.name]
                )
            properties[field.name] = field_schema.schema

        schema = {
            "type": "object",
            "name": type_.__name__,
            "properties": properties,
            "additionalProperties": False,
            "required": required,
        }
        if type_.__doc__:
            schema["description"] = type_.__doc__.strip()
        schema = Schema(schema, type_)
        return schema

    def from_json(
        self,
        item: ExternalItemType,
        context: SchemaContext,
        path: str,
        ref_schemas: Dict[str, Schema],
    ) -> Optional[Schema]:
        type_ = item.get("type")
        name = item.get("name")
        properties = item.get("properties")
        additional_properties = item.get("additionalProperties")
        if (
            type_ != "object"
            or not name
            or properties is None
            or additional_properties is not False
        ):
            return
        # noinspection PyTypeChecker
        ref_schema = Schema(
            {"$ref": path}, name
        )  # Having the name here implies that this will be a future annotation
        ref_schemas[path] = ref_schema
        params = {}
        annotations = {}
        params_with_default = {}

        for field_name, field_item in item.get("properties").items():
            field_schema = context.schema_from_json(
                field_item, f"{path}/{field_name}", ref_schemas
            )
            default = field_item.get("default", dataclasses.MISSING)
            if default is not dataclasses.MISSING:
                default = context.marshaller_context.load(
                    field_schema.python_type, default
                )
                p = params_with_default
            else:
                p = params
            field = dataclasses.field(
                default=default, metadata={"schemey": field_schema}
            )
            field.name = field_name
            field.type = field_schema.python_type
            annotations[field_name] = field.type
            p[field_name] = field
        params.update(**params_with_default)
        params["__annotations__"] = {
            k: annotations[k] for k in params
        }  # dataclasses uses annotation order...
        description = item.get("description")
        if description:
            params["__doc__"] = description
        type_ = dataclasses.dataclass(type(name, (), params))

        # Circular references are resolved at the end once we have a type with which to resolve them
        _resolve_futures(type_, name, type_)
        return Schema(item, type_)


def _resolve_futures(type_: Type, replace_name: str, replace_type: Type):
    if type_ == replace_name:
        return replace_type
    if typing_inspect.is_forward_ref(type_):
        import_name = typing_inspect.get_forward_arg(type_)
        if replace_name == import_name:
            return replace_type
    args = typing_inspect.get_args(type_)
    if args:
        args = tuple(_resolve_futures(a, replace_name, replace_type) for a in args)
        return type_.copy_with(args)
    if dataclasses.is_dataclass(type_):
        # noinspection PyDataclass
        for field in dataclasses.fields(type_):
            field.type = _resolve_futures(field.type, replace_name, replace_type)
    return type_



from __future__ import annotations

from dataclasses import dataclass, field
from typing import List, Dict, Type, Optional

from marshy import get_default_context
from marshy.marshaller_context import MarshallerContext
from marshy.types import ExternalItemType

from schemey.schema import Schema

# Referencing by name prevents circular reference
SchemaFactoryABC_ = "schemey.factory.schema_factory_abc.SchemaFactoryABC"


@dataclass
class SchemaContext:
    factories: List[SchemaFactoryABC_] = field(default_factory=list)
    schemas_by_type: Dict[Type, Schema] = field(default_factory=dict)
    marshaller_context: MarshallerContext = field(default_factory=get_default_context)

    def register_factory(self, schema_factory: SchemaFactoryABC_):
        self.factories.append(schema_factory)
        self.factories.sort(reverse=True)

    def schema_from_type(self, type_, path: Optional[str] = "#") -> Schema:
        schema = self.schemas_by_type.get(type_)
        if schema:
            return schema
        for factory in self.factories:
            schema = factory.from_type(type_, self, path)
            if schema:
                self.schemas_by_type[type_] = schema
                return schema
        raise ValueError(f"no_schema_for_type:{type_}")

    def schema_from_json(
        self,
        item: ExternalItemType,
        path: str = "#",
        ref_schemas: Optional[Dict[str, Schema]] = None,
    ) -> Schema:
        if ref_schemas is None:
            ref_schemas = {}
        for factory in self.factories:
            schema = factory.from_json(item, self, path, ref_schemas)
            if schema:
                return schema
        raise ValueError(f"no_schema_for_json:{item}")



from dataclasses import dataclass
from typing import Generic, TypeVar, Iterator, Optional, Type

from jsonschema import ValidationError
from marshy.marshaller.marshaller_abc import MarshallerABC
from marshy.types import ExternalItemType

from schemey import Schema, SchemaContext, get_default_schema_context
from schemey.validated import validated

T = TypeVar("T")


@dataclass
class Validator(Generic[T]):
    schema: Schema
    marshaller: MarshallerABC[T]

    def iter_errors(self, obj: T) -> Iterator[ValidationError]:
        item = self.marshaller.dump(obj)
        yield from self.schema.iter_errors(item)

    def validate(self, obj: T):
        item = self.marshaller.dump(obj)
        self.schema.validate(item)

    @property
    def json_schema(self) -> ExternalItemType:
        return self.schema.schema

    @property
    def python_type(self) -> Type[T]:
        # noinspection PyTypeChecker
        return self.schema.python_type

    @property
    def validated_type(self) -> Type[T]:
        # noinspection PyTypeChecker
        return validated(self.schema.python_type)


def validator_from_type(type_, context: Optional[SchemaContext] = None) -> Validator:
    if context is None:
        context = get_default_schema_context()
    return Validator(
        schema=context.schema_from_type(type_),
        marshaller=context.marshaller_context.get_marshaller(type_),
    )


def validator_from_json(
    item: ExternalItemType, context: Optional[SchemaContext] = None
) -> Validator:
    if context is None:
        context = get_default_schema_context()
    schema = context.schema_from_json(item)
    return Validator(
        schema=schema,
        marshaller=context.marshaller_context.get_marshaller(schema.python_type),
    )



from dataclasses import dataclass
from typing import Dict, Optional, Type

from marshy.types import ExternalItemType, ExternalType

from schemey import SchemaContext, Schema
from schemey.factory.schema_factory_abc import SchemaFactoryABC


@dataclass
class ExternalTypeFactory(SchemaFactoryABC):
    priority: int = 120

    def from_type(
        self, type_: Type, context: SchemaContext, path: str
    ) -> Optional[Schema]:
        if type_ == ExternalItemType:
            return Schema({"type": "object", "additionalProperties": True}, type_)
        elif type_ == ExternalType:
            return Schema({}, type_)

    def from_json(
        self,
        item: ExternalItemType,
        context: SchemaContext,
        path: str,
        ref_schemas: Dict[str, Schema],
    ) -> Optional[Schema]:
        if item == {}:
            return Schema(item, ExternalType)
        type_ = item.get("type")
        properties = item.get("properties")
        additional_properties = item.get("additionalProperties")
        if type_ == "object" and not properties and additional_properties is True:
            return Schema(item, ExternalItemType)



import importlib
import pkgutil
from typing import Optional, Type

from marshy import get_default_context
from marshy.marshaller_context import MarshallerContext
from marshy.types import ExternalItemType

from schemey.schema import Schema
from schemey.schema_context import SchemaContext

_default_context = None
CONFIG_MODULE_PREFIX = "schemey_config_"


def get_default_schema_context() -> SchemaContext:
    global _default_context
    if not _default_context:
        _default_context = new_default_schema_context()
    return _default_context


def new_default_schema_context(
    marshaller_context: Optional[MarshallerContext] = None,
) -> SchemaContext:
    if marshaller_context is None:
        marshaller_context = get_default_context()
    context = SchemaContext(marshaller_context=marshaller_context)
    # Set up context based on naming convention
    module_info = (
        m for m in pkgutil.iter_modules() if m.name.startswith(CONFIG_MODULE_PREFIX)
    )
    modules = [importlib.import_module(m.name) for m in module_info]
    modules.sort(key=lambda m: m.priority, reverse=True)
    for m in modules:
        getattr(m, "configure")(context)
    return context


def schema_from_type(type_: Type) -> Schema:
    return get_default_schema_context().schema_from_type(type_)


def schema_from_json(item: ExternalItemType) -> Schema:
    return get_default_schema_context().schema_from_json(item)



from abc import abstractmethod, ABC
from functools import total_ordering
from typing import Type, Optional, Dict

from marshy.types import ExternalItemType

from schemey.schema import Schema
from schemey.schema_context import SchemaContext


@total_ordering
class SchemaFactoryABC(ABC):
    @property
    def priority(self):
        return 100

    @abstractmethod
    def from_type(
        self, type_: Type, context: SchemaContext, path: str
    ) -> Optional[Schema]:
        """
        Create a schema for the type given, or return None if that was not possible
        """

    @abstractmethod
    def from_json(
        self,
        item: ExternalItemType,
        context: SchemaContext,
        path: str,
        ref_schemas: Dict[str, Schema],
    ) -> Optional[Schema]:
        """
        Create a schema for the type given, or return None if that was not possible
        """

    def __lt__(self, other):
        return self.priority < getattr(other, "priority", None)



from dataclasses import dataclass
from datetime import datetime
from typing import Optional, Type, Dict

from marshy.types import ExternalItemType

from schemey.factory.schema_factory_abc import SchemaFactoryABC
from schemey.schema import Schema, datetime_schema
from schemey.schema_context import SchemaContext
from schemey.string_format import StringFormat


@dataclass
class DatetimeFactory(SchemaFactoryABC):
    priority: int = 210

    def from_type(
        self, type_: Type, context: SchemaContext, path: str
    ) -> Optional[Schema]:
        if type_ is datetime:
            return datetime_schema()

    def from_json(
        self,
        item: ExternalItemType,
        context: SchemaContext,
        path: str,
        ref_schemas: Dict[str, Schema],
    ) -> Optional[Schema]:
        if (
            item.get("type") == "string"
            and item.get("format") == StringFormat.DATE_TIME.value
        ):
            return Schema(item, datetime)



from dataclasses import dataclass, fields
from typing import Optional, Type, TypeVar

from schemey import get_default_schema_context, SchemaContext, Schema

T = TypeVar("T")


class _ValidatingInstanceWrapper:
    def __init__(self, __wrapped__, __schemey_validators__):
        object.__setattr__(self, "__wrapped__", __wrapped__)
        object.__setattr__(self, "__schemey_validators__", __schemey_validators__)

    def __getattr__(self, item):
        wrapped = object.__getattribute__(self, "__wrapped__")
        return getattr(wrapped, item)

    def __setattr__(self, key, value):
        validators = object.__getattribute__(self, "__schemey_validators__")
        validator = validators.get(key)
        if validator:
            validator.validate(value)
        wrapped = object.__getattribute__(self, "__wrapped__")
        return setattr(wrapped, key, value)

    def __repr__(self):
        wrapped = object.__getattribute__(self, "__wrapped__")
        return wrapped.__repr__()

    def __eq__(self, other):
        wrapped = object.__getattribute__(self, "__wrapped__")
        if isinstance(other, _ValidatingInstanceWrapper):
            other = object.__getattribute__(other, "__wrapped__")
        return wrapped.__eq__(other)


class _ValidatingClassWrapper:
    def __init__(self, __wrapped__, __schemey_validators__=None):
        object.__setattr__(self, "__wrapped__", __wrapped__)
        object.__setattr__(self, "__schemey_validators__", __schemey_validators__)

    def __getattr__(self, item):
        wrapped = object.__getattribute__(self, "__wrapped__")
        return getattr(wrapped, item)

    def __call__(self, *args, **kwargs):
        wrapped_class = object.__getattribute__(self, "__wrapped__")
        to_wrap = wrapped_class.__call__(*args, **kwargs)
        validators = object.__getattribute__(self, "__schemey_validators__")
        for name, property_validator in validators.items():
            value = getattr(to_wrap, name)
            property_validator.validate(value)
        return _ValidatingInstanceWrapper(to_wrap, validators)


def validated(cls: Type[T], schema_context: Optional[SchemaContext] = None):
    if schema_context is None:
        schema_context = get_default_schema_context()

    def wrap(cls_):
        from schemey.validator import Validator

        type_ = dataclass(cls_)
        marshaller_context = schema_context.marshaller_context
        schema = schema_context.schema_from_type(type_)
        properties = schema.schema["properties"]
        validators = {}
        for field in fields(type_):
            prop = properties.get(field.name)
            if prop:
                schema = Schema(prop, field.type)
                marshaller = marshaller_context.get_marshaller(field.type)
                validators[field.name] = Validator(schema, marshaller)

        return _ValidatingClassWrapper(type_, validators)

    return wrap if cls is None else wrap(cls)



import setuptools

from schemey.version import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="schemey",
    version=__version__,
    author="Tim O'Farrell",
    author_email="tofarr@gmail.com",
    description="Convention over configuration Object Schemas for python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tofarr/schemey",
    packages=setuptools.find_packages(exclude=("tests", "tests.*")),
    install_requires=["marshy~=2.4", "jsonschema~=4.8"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from typing import Type, Optional, Tuple, Dict

import typing_inspect
from marshy.types import ExternalItemType

from schemey.factory.schema_factory_abc import SchemaFactoryABC
from schemey.schema import Schema
from schemey.schema_context import SchemaContext


class TupleSchemaFactory(SchemaFactoryABC):
    priority: int = 110

    def from_type(
        self, type_: Type, context: SchemaContext, path: str
    ) -> Optional[Schema]:
        origin = typing_inspect.get_origin(type_)
        if origin is tuple:
            args = typing_inspect.get_args(type_)
            if len(args) == 2 and args[1] is Ellipsis:
                return
            schema = {
                "type": "array",
                "prefixItems": [
                    context.schema_from_type(a, f"{path}/prefixItems/{i}").schema
                    for i, a in enumerate(args)
                ],
                "items": False,
            }
            return Schema(schema, type_)

    def from_json(
        self,
        item: ExternalItemType,
        context: SchemaContext,
        path: str,
        ref_schemas: Dict[str, Schema],
    ) -> Optional[Schema]:
        type_ = item.get("type")
        prefix_items = item.get("prefixItems")
        if type_ == "array" and prefix_items and item.get("items") is False:
            args = tuple(
                context.schema_from_json(
                    a, f"{path}/prefixItems/{i}", ref_schemas
                ).python_type
                for i, a in enumerate(prefix_items)
            )
            python_type = Tuple[args]
            return Schema(item, python_type)



from typing import Optional, Type, Dict

from marshy.types import ExternalItemType

from schemey.factory.schema_factory_abc import SchemaFactoryABC
from schemey.schema import Schema
from schemey.schema_context import SchemaContext

SCHEMA_FACTORY = "__schema_factory__"


class FactorySchemaFactory(SchemaFactoryABC):
    priority: int = 110

    def from_type(
        self, type_: Type, context: SchemaContext, path: str
    ) -> Optional[Schema]:
        factory = getattr(type_, SCHEMA_FACTORY, None)
        if factory is not None:
            return factory(context, path)

    def from_json(
        self,
        item: ExternalItemType,
        context: SchemaContext,
        path: str,
        ref_schemas: Dict[str, Schema],
    ) -> Optional[Schema]:
        """No implementation"""
